from __future__ import annotations

__author__: str
__license__: str
__version__: str

from rapidfuzz import (
    distance as distance,
    fuzz as fuzz,
    process as process,
    utils as utils,
)
